/*      X Window Video Capture Testing Tool
 *
 *      For testing a Video for Linux Two video capture driver
 *
 *      This program was written by Olivier Carmona, based on Bill Dirks.
 *      This program is in the public domain.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/mman.h>
#include <errno.h>

#include <linux/fs.h>
#include <linux/kernel.h>

#include <videodev.h>
#include <meteor_v4l2.h> // must follow videodev.h

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Porthole.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

/*  This program does not properly detect when the X Windows desktop is
 *  32 bits per pixel. (Some X functions use a value of 24 for the depth
 *  when it is really 32 for reasons I don't understand.) As a workaround,
 *  if you are running a 32-bpp X Windows screen, use -b on the command
 *  line, or uncomment this line:  
#define X_DEPTH 32
*/

/*  Set these according to your capture device set-up */
#define VID_DEVICE 	"/dev/video0" /* Device Name */
#define VID_PIX_FIELD 	V4L2_FIELD_TOP /* Video Output Format */
#define VID_STD	  	"PAL" /* "NTSC"  Default Standard */
#define VID_WIDTH  	384 /* 640  Default Width */
#define VID_HEIGHT 	288 /* 240  Default Height */
#define VID_PORT  	0 /* Video Input Port */
#define VID_STREAMING 	1 /* Streaming mode boolean */
#define STREAMBUFS	4 /* Number of streaming buffer */

/* Device Catpure Objects */
typedef struct tag_vimage
  {
    struct v4l2_buffer vidbuf;
    char *data;
    int   length;
  }
VIMAGE;

typedef struct
{
 int fd_video;
 int width;
 int height;
 int depth;
 int pixelformat;
 VIMAGE vimage[STREAMBUFS];
}capture_device;

/* Windows Display Objects */
typedef struct capture_window
{
 XtAppContext *app;
 Widget       w_toplevel;
 Widget       w_topbox;
 Widget       w_menubox;
 Widget       w_quit;
 Widget       w_viewport;
 Widget       w_stdbutton;
 Widget       w_stdmenu;
 Widget       w_sizebutton;
 Widget       w_sizemenu;
 GC           imagegc;
 int          x_depth;
}capture_window;

/* Standard Menu Objects */
typedef struct 
{
  Widget w_item;
  struct v4l2_standard std;
  capture_window *capt_w;
  capture_device *capt_d;
}
StdMenuItem;

/* Size Menu Objects */
struct
  {
    int width, height;
  }
capturesize[] =
{
  { 160, 120 } ,
  { 192, 144 } ,
  { 320, 240 } ,
  { 384, 288 } ,
  { 640, 480 } ,
  { 768, 576 } ,
};
 
typedef struct
{
  int index;
  capture_device *capt_d;
}
SzMenuItem;


/***************************/
/* Device Catpure Routines */
/***************************/

int capture_set_fmt (capture_device * pcapt_d)
{
  int fbf = V4L2_PIX_FMT_RGB565;

  if (pcapt_d->depth == 15)
    fbf = V4L2_PIX_FMT_RGB555;
  else if (pcapt_d->depth == 16)
    fbf = V4L2_PIX_FMT_RGB565;
  else if (pcapt_d->depth == 24)
    fbf = V4L2_PIX_FMT_BGR24;
  else if (pcapt_d->depth == 32)
    fbf = V4L2_PIX_FMT_BGR32;
  else
    {
      printf ("Unrecognized display depth. You may get an "
	      "X Windows error.\n");
    }
    
  return fbf;
}

#ifdef METEOR_V4L2_PERFORMANCE
int
capture_perf (capture_device * pcapt_d)
{
 struct v4l2_performance perf;
 
 if (ioctl (pcapt_d->fd_video, VIDIOC_G_PERF, &perf) != 0)
   printf ("G_PERF returned error code %d\n", errno);
 else
   printf ("Captured %d frames and lost %d frames\n", perf.frames,
		perf.framesdropped);
 return (perf.frames);
}
#endif // METEOR_V4L2_PERFORMANCE

void
capture_end (capture_device * pcapt_d)
{
 close (pcapt_d->fd_video);
}

void
capture_std (capture_device * pcapt_d, char *standard)
{
 int err;
 struct v4l2_standard std;
 
// v4l2_std NTSC to Standard
 err = ioctl (pcapt_d->fd_video, VIDIOC_S_STD, &std);
 if (err)
   {
     perror ("S_STD in capture_std");
   }
}

void
capture_size (capture_device * pcapt_d, int width, int height)
{
 struct v4l2_format fmt;
 int err;

 fmt.fmt.pix.width = width;
 fmt.fmt.pix.height = height;
 err = ioctl (pcapt_d->fd_video, VIDIOC_S_FMT, &fmt);
 if (err)
   {
     perror ("S_FMT in capture_size");
   }
}

void
capture_stop (capture_device * pcapt_d, int streaming)
{
  int i, err;

  if (pcapt_d->fd_video >= 0)
    {
      if (streaming)
        {
          i = V4L2_BUF_TYPE_VIDEO_CAPTURE;
          err = ioctl (pcapt_d->fd_video, VIDIOC_STREAMOFF, &i);
          for (i = 0; i < STREAMBUFS; ++i)
  	    {
	      if (pcapt_d->vimage[i].data)
	        munmap (pcapt_d->vimage[i].data, pcapt_d->vimage[i].length);
	      pcapt_d->vimage[i].data = NULL;
	    }
	}
      else
        {
          if (pcapt_d->vimage[0].data)
  	    free (pcapt_d->vimage[0].data);
          pcapt_d->vimage[0].data = NULL;
	}
    }
}

int
capture_start (capture_device * pcapt_d, int streaming)
{
 struct v4l2_requestbuffers req;
 int i, err;

 if (streaming)
   {
     /* Ask Video Device for Buffers */
     req.count = STREAMBUFS;
     req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
     err = ioctl (pcapt_d->fd_video, VIDIOC_REQBUFS, &req);
     if (err < 0 || req.count < 1)
       {
         perror ("REQBUFS in capture _start\n");
         return 0;
       }
    
     /* Query each buffer and map it to the video device */
     for (i = 0; i < req.count; ++i)
       {
         struct v4l2_buffer *vidbuf = &pcapt_d->vimage[i].vidbuf;
         vidbuf->index = i;
         vidbuf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
         err = ioctl (pcapt_d->fd_video, VIDIOC_QUERYBUF, vidbuf);
         if (err < 0)
   	   {
	     perror ("QUERYBUF in capture_start");
	     return 0;
	   }
	 fprintf(stderr, "Buffer Nr. %d Lenght %d Offset %d\n",
	    i, vidbuf->length, vidbuf->m.offset );
	   
         pcapt_d->vimage[i].length = 0;
         pcapt_d->vimage[i].data = mmap (0, vidbuf->length, 
	           PROT_READ|PROT_WRITE, MAP_PRIVATE/*MAP_SHARED*/,
	           pcapt_d->fd_video, vidbuf->m.offset);
         if ((int) pcapt_d->vimage[i].data == -1)
	   {
	     perror ("mmap() in capture_start");
	     return 0;
   	   }
         pcapt_d->vimage[i].length = vidbuf->length; 

	 err = ioctl (pcapt_d->fd_video, VIDIOC_QBUF, vidbuf);
         if (err)
           {
	     perror ("QBUF in capture_start");
	     return 0;
           }
       }
       
     /* Set video stream capture on */
     err = ioctl (pcapt_d->fd_video, VIDIOC_STREAMON, &pcapt_d->vimage[0].vidbuf.type);
     if (err)
       perror ("STREAMON in capture_start");
   }
 else
    {
      /* Alloc one Buffer for one image */
      int sizeimg = pcapt_d->width*pcapt_d->height*pcapt_d->depth/8;
      
      pcapt_d->vimage[0].data = malloc (sizeimg);
      if (pcapt_d->vimage[0].data == NULL)
        {
          printf ("malloc(%d) failed in capture_start\n", sizeimg);
          return 0;
        }
    }

  return 1;
}


int
capture_init (capture_device * pcapt_d, char *my_device, int input, 
              unsigned int format, int width, int height, int depth)
{
  int err;
  struct v4l2_capability cap;
  struct v4l2_streamparm parm;
  struct v4l2_standard std;
  struct v4l2_format fmt;

  /* Open Video Device */
  pcapt_d->fd_video = open (my_device, O_RDONLY);
  if (pcapt_d->fd_video < 0)
    {
      printf ("No video device \"%s\"\n", my_device);
      return 0;
    }

  /* Querry Video Device Capabilities */
  err = ioctl (pcapt_d->fd_video, VIDIOC_QUERYCAP, &cap);
  if (err)
    {
      perror ("QUERYCAP in capture_init");
      return 0;
    }
  if (cap.capabilities != V4L2_CAP_VIDEO_CAPTURE)
    {
      printf ("Not a capture device.\n");
      return 0;
    }
  /*
  if (!(cap.flags & V4L2_FLAG_READ))
    {
      printf ("Device does not support the read() call.\n");
    }
  if (!(cap.flags & V4L2_FLAG_STREAMING))
    {
      printf ("Device does not support streaming capture.\n");
    }
  */

  /* Set Video Input Port */
  err = ioctl (pcapt_d->fd_video, VIDIOC_S_INPUT, &input);
  if (err)
    perror ("S_INPUT in capture_init");

  /* Set Video Parameters */
  parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_PARM, &parm);
  if (err)
    perror ("G_PARM in capture_init");

  printf("Time per frame %d/%d", 
		  parm.parm.capture.timeperframe.numerator,
		  parm.parm.capture.timeperframe.denominator );
  
  err = ioctl (pcapt_d->fd_video, VIDIOC_S_PARM, &parm);
  if (err) {
    perror ("S_PARM in capture_init");
    // exit(1);
  }
    
  /* Get Video Standard */
  std.index = 0;
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_STD, &std);
  if (err)
    perror ("G_STD in capture_init");

  /* Set Video Format */
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  err = ioctl (pcapt_d->fd_video, VIDIOC_G_FMT, &fmt);
  if (err)
    perror ("G_FMT in capture_init");

  if (width == 0)
    width = VID_WIDTH;
  fprintf(stderr, "Switching width from %d to %d\n",
			fmt.fmt.pix.width, width); 
  fmt.fmt.pix.width = width;
    
  if (height == 0) 
    height = VID_HEIGHT;
  fprintf(stderr, "Switching height from %d to %d\n",
    fmt.fmt.pix.height, height);
  fmt.fmt.pix.height = height;
  
  if (depth) {
    fprintf(stderr, "depth %d \n", depth );
    // should use sizeimage ???
    // fmt.fmt.pix.depth = depth;
    // fmt.fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
  } 
    
  fmt.fmt.pix.field = VID_PIX_FIELD;
#ifdef VID_PIXELFORMAT
  fmt.fmt.pix.pixelformat = VID_PIXELFORMAT;
#else
  fmt.fmt.pix.pixelformat = format;
#endif

  pcapt_d->width = fmt.fmt.pix.width;
  pcapt_d->height = fmt.fmt.pix.height;
  // pcapt_d->depth = fmt.fmt.pix.depth;
  pcapt_d->pixelformat = fmt.fmt.pix.pixelformat;
  
  err = ioctl (pcapt_d->fd_video, VIDIOC_S_FMT, &fmt);
  if (err)
    perror ("S_FMT in capture_init");

  
  return (1);
}


int
capture_dq (capture_device * pcapt_d,
	    struct v4l2_buffer *pbuf)
{
  int err, n;
  fd_set rdset;
  struct timeval timeout;

  FD_ZERO (&rdset);
  FD_SET (pcapt_d->fd_video, &rdset);

  timeout.tv_sec = 1;
  timeout.tv_usec = 0;

  n = select (pcapt_d->fd_video + 1, &rdset, NULL, NULL, &timeout);
  if (n == -1)
    fprintf (stderr, "select error.\n");
  else if (n == 0)
    fprintf (stderr, "select timeout\n");
  else if (FD_ISSET (pcapt_d->fd_video, &rdset))
    {
      pbuf->type = pcapt_d->vimage[0].vidbuf.type;
      err = ioctl (pcapt_d->fd_video, VIDIOC_DQBUF, pbuf);
      if (err)
	{
	  perror ("DQBUF in capture_dq");
	  return 0;
	}

      return 1;
    }

  return 0;
}

void
capture_q (capture_device * pcapt_d,
	   struct v4l2_buffer *pbuf)
{
  int err;

  err = ioctl (pcapt_d->fd_video, VIDIOC_QBUF, pbuf);
  if (err)
    perror ("QBUF in capture_q");
}

int
capture_read (capture_device * pcapt_d, int sz_img)
{
  int err;
  
  err = read (pcapt_d->fd_video, pcapt_d->vimage[0].data, sz_img);
  if (err < 0)
    {
      perror ("READ in capture_read");
      return 0;
    }
  
  return 1;
}


/****************************/
/* Windows Display Routines */
/****************************/

int display_get_depth (capture_window * pcapt_w)
{
  if (pcapt_w->x_depth == 0)
    {
#ifdef X_DEPTH
      pcapt_w->x_depth = X_DEPTH;
#else
      Display *d = XtDisplay (pcapt_w->w_viewport);
      Window w = DefaultRootWindow (XtDisplay (pcapt_w->w_viewport));

      XWindowAttributes wa;
      XGetWindowAttributes (d, w, &wa);
      pcapt_w->x_depth = wa.depth;
#endif
    }
  return (pcapt_w->x_depth);
}

int
display_start (capture_window * pcapt_w, int width, int height)
{
  Arg arg[2];

  XtSetArg (arg[0], XtNwidth, width);
  XtSetArg (arg[1], XtNheight, height);
  XtSetValues (pcapt_w->w_viewport, arg, 2);
  XtRealizeWidget (pcapt_w->w_toplevel);

  return 1;
}

void
display_stop (capture_window * pcapt_w, XImage *XImgTmp)
{
 if (XImgTmp)
// memory leak, but XDestroyImage always segfaults...
//          XDestroyImage(XImgTmp);
   XImgTmp = NULL;
   
 return;
}

void
quit_callback (Widget w, XtPointer client, XtPointer call)
{
  capture_device *pcapt_d = (capture_device *) client;

  if (pcapt_d->fd_video >= 0)
    {
      #ifdef METEOR_V4L2_PERFORMANCE
        capture_perf(pcapt_d);
      #endif
      capture_stop (pcapt_d, VID_STREAMING);
      capture_end (pcapt_d); 
    }
  exit (0);
}

void
std_menu_callback (Widget w, XtPointer client, XtPointer call)
{
  StdMenuItem *smi = (StdMenuItem *) client;
  capture_device *pcapt_d = smi->capt_d;
  capture_window *pcapt_w = smi->capt_w;

  capture_stop (pcapt_d, VID_STREAMING);
  capture_std (pcapt_d, VID_STD);
  capture_start (pcapt_d, VID_STREAMING);
  display_start (pcapt_w, pcapt_d->width, pcapt_d->height);
}

void
size_menu_callback (Widget w, XtPointer client, XtPointer call)
{
  SzMenuItem *szi = (SzMenuItem *) client;
  int i = szi->index;
  capture_device *pcapt_d = szi->capt_d;
  
  capture_stop (pcapt_d, VID_STREAMING);
  capture_size (pcapt_d, capturesize[i].width, capturesize[i].height);
  capture_start (pcapt_d, VID_STREAMING);
}

int
display_init (capture_window * pcapt_w, 
              capture_device * pcapt_d,
	      int argc, char *argv[], 
	      int width, int height)
{
  int err, i;

  pcapt_w->app = (XtAppContext *) malloc (sizeof (XtAppContext));
  pcapt_w->w_toplevel = XtAppInitialize (pcapt_w->app, "DTSS", NULL, 0,
					 &argc, argv, NULL, NULL, 0);
  pcapt_w->w_topbox = XtVaCreateManagedWidget ("topbox",
					       boxWidgetClass,
					       pcapt_w->w_toplevel,
					       XtNorientation,
					       XtorientVertical,
					       XtNborderWidth, 0, XtNhSpace,
					       0, XtNvSpace, 0, NULL);
  pcapt_w->w_menubox =
    XtVaCreateManagedWidget ("menubox", boxWidgetClass, pcapt_w->w_topbox,
			     XtNorientation, XtorientHorizontal,
			     XtNborderWidth, 0, XtNhSpace, 4, XtNvSpace, 1,
			     NULL);

  pcapt_w->w_stdbutton = XtVaCreateManagedWidget ("menubutton",
						  menuButtonWidgetClass,
						  pcapt_w->w_menubox,
						  XtNlabel, "Standard",
						  XtNmenuName, "standard",
						  XtNborderWidth, 0, NULL);
  pcapt_w->w_stdmenu = XtVaCreatePopupShell ("standard",
					     simpleMenuWidgetClass,
					     pcapt_w->w_stdbutton, NULL);


  for (i = 0;; ++i)
    {
      struct v4l2_enumstd estd;
      StdMenuItem *smi;

      estd.index = i;
      err = ioctl (pcapt_d->fd_video, VIDIOC_ENUMSTD, &estd);
      if (err)
	break;
      smi = (StdMenuItem *) malloc (sizeof (StdMenuItem));
      smi->std = estd.std;
      smi->capt_w = pcapt_w;
      smi->capt_d = pcapt_d;
      smi->w_item = XtVaCreateManagedWidget (estd.std.name,
					     smeBSBObjectClass,
					     pcapt_w->w_stdmenu, XtNlabel,
					     estd.std.name, NULL);
      XtAddCallback (smi->w_item, XtNcallback, std_menu_callback, smi);
    }

  pcapt_w->w_sizebutton = XtVaCreateManagedWidget ("menubutton",
						   menuButtonWidgetClass,
						   pcapt_w->w_menubox,
						   XtNlabel, "Size",
						   XtNmenuName, "size",
						   XtNborderWidth, 0, NULL);
  pcapt_w->w_sizemenu = XtVaCreatePopupShell ("size",
					      simpleMenuWidgetClass,
					      pcapt_w->w_sizebutton, NULL);
  for (i = 0; i < sizeof (capturesize) / sizeof (capturesize[0]); ++i)
    {
      char sizestr[32];
      SzMenuItem *szi;
      Widget w;

      szi = (SzMenuItem *) malloc (sizeof (SzMenuItem));
      szi->index = i;
      szi->capt_d = pcapt_d;
      sprintf (sizestr, "%d x %d", capturesize[i].width,
	       capturesize[i].height);
      w = XtVaCreateManagedWidget (sizestr,
				   smeBSBObjectClass, pcapt_w->w_sizemenu,
				   XtNlabel, sizestr, NULL);

      XtAddCallback (w, XtNcallback, size_menu_callback, (char *) szi);
    }

  pcapt_w->w_quit = XtVaCreateManagedWidget ("Quit!",
					     commandWidgetClass,
					     pcapt_w->w_menubox,
					     XtNborderWidth, 0, NULL);
  XtAddCallback (pcapt_w->w_quit, XtNcallback, quit_callback, pcapt_d);
  pcapt_w->w_viewport = XtVaCreateManagedWidget ("port",
						 portholeWidgetClass,
						 pcapt_w->w_topbox, XtNwidth,
						 width, XtNheight,
						 height, NULL);
  XtRealizeWidget (pcapt_w->w_toplevel);
  pcapt_w->imagegc = XCreateGC (XtDisplay (pcapt_w->w_viewport),
				XtWindow (pcapt_w->w_viewport), 0, NULL);
  if (pcapt_w->imagegc == NULL)
    {
      printf ("XCreateGC failed.\n");
      return 0;
    }

  pcapt_w->x_depth = 0;

  return 1;
}

XImage *
display_alloc (capture_window * pcapt_w, char *data, int width, int height)
{
 XImage *XImgTmp;

 XImgTmp = XCreateImage (XtDisplay (pcapt_w->w_viewport),
 	       DefaultVisual (XtDisplay (pcapt_w->w_viewport), 0),
	       (pcapt_w->x_depth == 32) ? 24 : pcapt_w->x_depth,
	       ZPixmap, 0, data, width, height, 16, 0);
	       
 if (XImgTmp == NULL)
   {
     printf ("No XImage\n");
   }
  
  return XImgTmp;
}

int
display_status (capture_window * pcapt_w)
{
  XEvent event;

  if (XtIsRealized (pcapt_w->w_viewport))
    {
      return 1;
    }
  else
    {
      XtAppPeekEvent (*pcapt_w->app, &event);
      return 0;
    }

}

void
display_refresh (capture_window * pcapt_w, 
                 XImage *XImgTmp, int width, int height)
{

  XPutImage (XtDisplay (pcapt_w->w_viewport),
	     XtWindow (pcapt_w->w_viewport),
	     pcapt_w->imagegc, XImgTmp, 0, 0, 0, 0, width, height);
}


void
display_wakeup (capture_window * pcapt_w)
{
  XEvent event;

  while (XtAppPending (*pcapt_w->app))
    {
      XtAppNextEvent (*pcapt_w->app, &event);
      switch (event.type)
	{
	  default:
	    break;
	}
      XtDispatchEvent (&event);
    }
}

int
main (int argc, char *argv[])
{
  capture_window capt_w;
  capture_device capt_d;
  XImage *capt_ximg[STREAMBUFS];
  unsigned int format = V4L2_PIX_FMT_BGR32;
  int argi;
  register int i;

  /* Video Parameters */
  char my_device[64];
  int input;

  /*    Put in the device node name */
  strcpy (my_device, VID_DEVICE);
  input = VID_PORT;

  /* Read arguments */
  for (argi = 1; argi < argc; ++argi)
    {
      if (strcmp (argv[argi], "-b") == 0)
	{
	  capt_w.x_depth = 32;
	  continue;
	}
      strcpy (my_device, argv[argi]);
    }

  capture_init (&capt_d, my_device, input, format, 0, 0, 0);
  
  display_init (&capt_w, &capt_d, argc, argv, capt_d.width, capt_d.height);
  
  display_get_depth (&capt_w);
  fprintf(stderr, "Display Depth %d \n", capt_d.depth );

  capture_start (&capt_d, VID_STREAMING);
  
  for (i=0; i < STREAMBUFS; i++)
    {
      capt_ximg[i] = display_alloc (&capt_w, capt_d.vimage[i].data, 
                                    capt_d.width, capt_d.height);
    }
  
  display_start (&capt_w, capt_d.width, capt_d.height);
    
  fprintf (stderr, "Capturing %dx%dx%d \"%4.4s\" images\n",
	  capt_d.width, capt_d.height, capt_d.depth,
	  (char *) &(capt_d.pixelformat));
  fprintf (stderr, "Images are %d bytes each\n", 
          capt_d.width * capt_d.height * capt_d.depth);

  for (;;)
    {
      struct v4l2_buffer tempbuf;

      if (display_status (&capt_w))
	{
	  if (capture_dq (&capt_d, &tempbuf))
	    {
	      display_refresh (&capt_w, capt_ximg[tempbuf.index], 
	        capt_d.width, capt_d.height);
  	      capture_q (&capt_d, &tempbuf);
	    }
	}

      display_wakeup (&capt_w);
    }

  return 0;
}

